;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    this file is part of CavityStuffer - a molecular polymer design tool
;;;                         -------------
;;; copyright (c) 1993 by Institute for Molecular Manufacturing, Palo Alto, CA
;;; copyright (c) 1994 and beyond: M. Krummenacker (KR) kr@n-a-n-o.com
;;;
;;; -This program is free software; you can redistribute it and/or modify
;;;   it under the terms of the GNU General Public License as published by
;;;   the Free Software Foundation.
;;; -This program is distributed in the hope that it will be useful, but WITHOUT
;;;   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;;;   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU GPL for more details.
;;; -You should have received a copy of the GNU General Public License along with
;;;   this program (the file "COPYING-2.0"); if not, write to the Free Software
;;;   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Historical Note: the first version of this file was made by K. E. Drexler.
;;;                  subsequent modifications by KR.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package "CAVSTUF")

;;; This defines several functions that share a single
;;; array-based representation of atoms in space. Array elements
;;; correspond to cubical volumes of width max-clash-distance,
;;; chosen such that two atoms separated by a full block cannot
;;; clash. Each array element contains a list of the atoms in its
;;; cubical volume. The size of the array is large enough
;;; to include all of a set of bounding points, given as a list of 
;;; (x y z) coordinate lists. The clash distances are computed from
;;; a list of minimum atomic radii, given in sequence that
;;; corresponds to the atom type numbers (starting with type 0).
;;;
;;; Functions: (out-of-date (kr 93-8-22))
;;;
;;; (insert-in-clash-array atm)       args: atm
;;;                                returns: nothing
;;; (delete-from-clash-array atm)     args: atm
;;;                                returns: nothing
;;; (clash-p atm ignore-list)         args: atm, list of atms
;;;                                returns: t or nil
;;; (initialize-clash-array atm-list) args: list of atms
;;;                                returns: nothing
;;; (reset-clash-array)               args: none
;;;                                returns: nothing
;;; (extract-clash-array)             args: none
;;;                                returns: clash-array

(defun set-clash-environment (bounding-points-list min-radii-list)
  (let* ; establishes the lexical environment for the clash functions
    ;; /max-clash is reciprocal of maximum distance for a clash.
    ((max-radius (apply #'max min-radii-list))
     (/max-clash (/ 1.0 (* 2.0 max-radius)))
     (radius (apply #'vector min-radii-list))
     (x-low  (apply #'min (mapcar #'first  bounding-points-list)))
     (x-high (apply #'max (mapcar #'first  bounding-points-list)))
     (y-low  (apply #'min (mapcar #'second bounding-points-list)))
     (y-high (apply #'max (mapcar #'second bounding-points-list)))
     (z-low  (apply #'min (mapcar #'third  bounding-points-list)))
     (z-high (apply #'max (mapcar #'third  bounding-points-list)))
     (x-dim  (ceiling (* (- x-high x-low) /max-clash)))
     (y-dim  (ceiling (* (- y-high y-low) /max-clash)))
     (z-dim  (ceiling (* (- z-high z-low) /max-clash)))
     (clash-array
      (make-array (list x-dim y-dim z-dim) :initial-element nil))
     (initial-clash-array nil)
     (x-array-lim (- x-dim 1))
     (y-array-lim (- y-dim 1))
     (z-array-lim (- z-dim 1))
     (square-of-min-distance
      (make-array (list (length min-radii-list)
                        (length min-radii-list))))
     )
    ;; initialize square-of-min-distance array
    (let ((n (length radius)))
      (dotimes (i n)
        (dotimes (j n)
          (setf (aref square-of-min-distance i j)
                (let ((rsum (+ (aref radius i)
                               (aref radius j))))
                  (* rsum rsum))))))
    
    ;;; Returns volume of clash-array space ***
    (defun get-clash-bounds ()
      (values x-low x-high y-low y-high z-low z-high))
    
    (defun insert-multiple-into-clash-array (atm-array)
      "atm-array
    Inserts all the atms of the atm-array into the clash-array"
      (dosequence (atm atm-array)
                  (insert-in-clash-array atm)))

    ;;; Places an atom in the clash-array.
    (defun insert-in-clash-array (atm)
      (push atm
            (aref clash-array
                  (floor (* (- (aref (atm-xyz atm) 0) x-low)
                            /max-clash))
                  (floor (* (- (aref (atm-xyz atm) 1) y-low)
                            /max-clash))
                  (floor (* (- (aref (atm-xyz atm) 2) z-low)
                            /max-clash))))
      ;; Maybe the following should be modified to implement
      ;;   the periodic boundary conditions right here
      (let ((image (atm-image atm)))
        (when image
          (insert-in-clash-array image)))
      (values))
    
    (defun delete-multiple-from-clash-array (atm-array)
      "atm-array
    Deletes all the atms of the atm-array from the clash-array"
      (dosequence (atm atm-array)
                  (delete-from-clash-array atm)))
    
    ;;; Deletes an atom from the clash-array.
    (defun delete-from-clash-array (atm)
      (delete-if
       #'(lambda (sample-atm) (eq sample-atm atm))
       (aref clash-array
             (floor (* (- (aref (atm-xyz atm) 0) x-low) /max-clash))
             (floor (* (- (aref (atm-xyz atm) 1) y-low) /max-clash))
             (floor (* (- (aref (atm-xyz atm) 2) z-low) /max-clash))))
      (let ((image (atm-image atm)))
        (when image
          (delete-from-clash-array image)))
      (values))
    
    (defun outside-boundaries-p (x y z r)
      "x y z r
     This primitive version implements a simple rectilinear box."
      (or (< x (+ x-low r))
          (> x (- x-high r))
          (< y (+ y-low r))
          (> y (- y-high r))
          (< z (+ z-low r))
          (> z (- z-high r))))

    (defun clash-multiple-p (atm-array ignore-list)
      "atm-array
       ignore-list
    Test all the atms of the atm-array for clash in the clash-array.
    Break off the testing as soon some clash is encountered."
      (some #'(lambda (atm)
     ;           (if                            ;* for diagnostics
                  (clash-p atm ignore-list)
     ;             (progn                       ;*
     ; (print "diagnostics-msg from clash-multiple-p:");*
     ; (print "  first clashing atm is:")       ;*
     ; (print atm)                              ;*
     ;               t)                         ;*
     ;             nil)                         ;*
                )
            atm-array))

    ;;; Returns t iff the given atom clashes with an existing
    ;;; member of the clash-array. Identifies the cell into which
    ;;; the given atom falls, then compares its squared-distance
    ;;; from atoms in this and adjacent (within-array) cells to
    ;;; the minimum permissible squared-distance for the two types.
    ;;; Ignores clashes with atoms on the ignore-list.
    ;;; Clash-p must be given an atom or image within the periodic
    ;;; boundaries.
    (defun clash-p (atm ignore-list)
      (let* ((x (aref (atm-xyz atm) 0))
             (y (aref (atm-xyz atm) 1))
             (z (aref (atm-xyz atm) 2))
             (type (atm-type atm))
             (r (aref radius type))
             (clash nil)) ; set to t if a clash is detected
        (if (outside-boundaries-p x y z r)
          (setf clash t)
          (let* ((middle-cell-x (floor (* (- x x-low) /max-clash)))
                 (middle-cell-y (floor (* (- y y-low) /max-clash)))
                 (middle-cell-z (floor (* (- z z-low) /max-clash)))
                 (low-cell-x (max 0 (- middle-cell-x 1)))
                 (low-cell-y (max 0 (- middle-cell-y 1)))
                 (low-cell-z (max 0 (- middle-cell-z 1)))
                 (high-cell-x (min x-array-lim (+ middle-cell-x 1)))
                 (high-cell-y (min y-array-lim (+ middle-cell-y 1)))
                 (high-cell-z (min z-array-lim (+ middle-cell-z 1))))
            (do ((i low-cell-x (+ 1 i)))
                ((or (> i high-cell-x) clash))
              (do ((j low-cell-y (+ 1 j)))
                  ((or (> j high-cell-y) clash))
                (do ((k low-cell-z (+ 1 k)))
                    ((or (> k high-cell-z) clash))
                  (dolist (array-atm (aref clash-array i j k))
                    (let ((max-current-clash
                           (+ r (aref radius (atm-type array-atm))))
                          (dx (- x (aref (atm-xyz array-atm) 0))))
                      ;; Structured for early rejection, some gain.
                      (when (< (abs dx) max-current-clash)
                        (let ((dy (- y
                                     (aref (atm-xyz array-atm) 1))))
                          (when (< (abs dy) max-current-clash)
                            (let
                              ((dz (- z
                                      (aref (atm-xyz array-atm) 2))))
                              (when
                                (and
                                 (< (abs dz) max-current-clash)
                                 (< (+ (* dx dx) (* dy dy) (* dz dz))
                                    (aref square-of-min-distance
                                          type
                                          (atm-type array-atm)))
                                 (not (find-if
                                       #'(lambda (a) (eq a array-atm))
                                       ignore-list)))
                                (setf clash t)
                                (return)))))))))))))
        clash))

    ;;; Returns the first atm from the clash-array with which the
    ;;;   supplied atm clashes.
    (defun diagnostic-clash (atm ignore-list)
      (let* ((x (aref (atm-xyz atm) 0))
             (y (aref (atm-xyz atm) 1))
             (z (aref (atm-xyz atm) 2))
             (type (atm-type atm))
             (r (aref radius type))
             (clash nil)) ; set to the clashing atm
                          ;   if a clash is detected
        (if (outside-boundaries-p x y z r)
          (setf clash t)
          (let* ((middle-cell-x (floor (* (- x x-low) /max-clash)))
                 (middle-cell-y (floor (* (- y y-low) /max-clash)))
                 (middle-cell-z (floor (* (- z z-low) /max-clash)))
                 (low-cell-x (max 0 (- middle-cell-x 1)))
                 (low-cell-y (max 0 (- middle-cell-y 1)))
                 (low-cell-z (max 0 (- middle-cell-z 1)))
                 (high-cell-x (min x-array-lim (+ middle-cell-x 1)))
                 (high-cell-y (min y-array-lim (+ middle-cell-y 1)))
                 (high-cell-z (min z-array-lim (+ middle-cell-z 1))))
            (do ((i low-cell-x (+ 1 i)))
                ((or (> i high-cell-x) clash))
              (do ((j low-cell-y (+ 1 j)))
                  ((or (> j high-cell-y) clash))
                (do ((k low-cell-z (+ 1 k)))
                    ((or (> k high-cell-z) clash))
                  (dolist (array-atm (aref clash-array i j k))
                    (let ((max-current-clash
                           (+ r (aref radius (atm-type array-atm))))
                          (dx (- x (aref (atm-xyz array-atm) 0))))
                      ;; Structured for early rejection, some gain.
                      (when (< (abs dx) max-current-clash)
                        (let ((dy (- y
                                     (aref (atm-xyz array-atm) 1))))
                          (when (< (abs dy) max-current-clash)
                            (let
                              ((dz (- z
                                      (aref (atm-xyz array-atm) 2))))
                              (when
                                (and
                                 (< (abs dz) max-current-clash)
                                 (< (+ (* dx dx) (* dy dy) (* dz dz))
                                    (aref square-of-min-distance
                                          type
                                          (atm-type array-atm)))
                                 (not (find-if
                                       #'(lambda (a) (eq a array-atm))
                                       ignore-list)))
                                (setf clash array-atm)
                                (return)))))))))))))
        clash))

    (flet
      (
       ;; Makes a copy of a clash-array with no shared structure.
       (copy-clash-array
        (old-array)
        (let ((new-array 
               (make-array (list x-dim y-dim z-dim)
                           :initial-element nil)))
          (dotimes (i x-dim)
            (dotimes (j y-dim)
              (dotimes (k z-dim)
                (setf (aref new-array i j k)
                      (copy-seq
                       (aref old-array i j k))))))
          new-array))
       )
      
      ;;; Places multiple atoms in the clash-array and establishes
      ;;; this state as the initial value restored by the
      ;;; reset-clash-array function. If never invoked, less space
      ;;; is consumed, because only one clash-array object exists.
      (defun initialize-clash-array (atm-list)
        (dolist (atm atm-list)
          (insert-in-clash-array atm))
        (setf initial-clash-array
              (copy-clash-array clash-array))
        (values))
      
      ;;; Discards the the clash-array, restoring an initialized
      ;;; state if there is one, else leaving it empty.
      (defun reset-clash-array ()
        (if (null initial-clash-array)
          (setf clash-array
                (make-array
                 (list x-dim y-dim z-dim) :initial-element nil))
          (setf clash-array
                (copy-clash-array initial-clash-array)))
        (values))

      ;;; Returns clash-array (for diagnostics).
      (defun extract-clash-array ()
        clash-array)

      )))

#|
;;; This function belongs with the moiety installation
;;; routines, and is used to initialize the moiety-clash-atms???
;;; list to include image atms; it is used here as a test.

(defun ignore-with-images (primary-atm-ignore-list)
  "primary-atm-ignore-list
Appends to the primary-atm-ignore-list all the images of these atms,
so they too will be ignored in the clash-array."
  (labels ((list-images
            (atm)
            (if (null (atm-image atm))
              (list atm)
              (cons atm (list-images (atm-image atm))))))
    (apply #'append (mapcar #'(lambda (atm) (list-images atm))
                            primary-atm-ignore-list))))

(let ((atm1 (make-atm :xyz (vector 1 1 1) :type 1))
      (atm2 (make-atm :xyz (vector 5 1 1) :type 1))
      (atm3 (make-atm :xyz (vector 5 1 1) :type 1)))
  (set-clash-environment '((0 0 0) (10 10 10))
                         '(0.0 1.0 1.5 1.9))
  (setf (atm-image atm1) atm2)
  (insert-in-clash-array atm1)
  (clash-p atm3 (ignore-with-images (list atm1))))

(let ((result 0))
  (set-clash-environment '((0 0 0) (10 10 10))
                         '(0.0 1.5 1.9))
  (dotimes (i 100) ;; 1:5  2:10:11  5:20  10:50:15  20:70 50:95:19
    (insert-in-clash-array (make-atm :xyz
                                     (vector (random 10.0)
                                             (random 10.0)
                                             (random 10.0))
                                     :type 1)))
  (time (dotimes (i 10000)
          (when (clash-p (make-atm :xyz
                                   (vector (random 10.0)
                                           (random 10.0)
                                           (random 10.0))
                                   :type 0)
                         nil)
            (incf result))))
  (values result))

|# 
