;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    this file is part of CavityStuffer - a molecular polymer design tool
;;;                         -------------
;;; copyright (c) 1993 by Institute for Molecular Manufacturing, Palo Alto, CA
;;; copyright (c) 1994 and beyond: M. Krummenacker (KR) kr@n-a-n-o.com
;;;
;;; -This program is free software; you can redistribute it and/or modify
;;;   it under the terms of the GNU General Public License as published by
;;;   the Free Software Foundation.
;;; -This program is distributed in the hope that it will be useful, but WITHOUT
;;;   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;;;   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU GPL for more details.
;;; -You should have received a copy of the GNU General Public License along with
;;;   this program (the file "COPYING-2.0"); if not, write to the Free Software
;;;   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Historical Note: an initial version of this file was made by K. E. Drexler.
;;;                  subsequent modifications by KR.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package "CAVSTUF")


;;; Revisions(by KED 93/1/22): There is an option
;;; to include atoms in a PDB file that help define the structure
;;; (including the coordinate system) but that are omitted in the
;;; moiety-type structure. This enables, for example, the entry
;;; of a moiety with fewer than three atoms, and the entry of a
;;; methyl moiety such that none of its cap atoms are used as
;;; coordinate defining atoms -- which would otherwise lead to
;;; difficulties in representing the relationship between a methyl
;;; moiety and its child moieties.

;;; NOTE on used data formats:
;;; -------------------------
;;; The output from parsing action (by the function
;;; read&parse-all-lines-of-a-file) on a molecule-describing file
;;; of a format like PDB, is called raw-format, and consists of
;;; a list of two vectors, the first containing parsed-atms,
;;; the second parsed-bond-starlets.
;;; - A parsed-atm is a list, the first element being an integer
;;;   which denotes the (unique) number of the atom in the molecule,
;;;   and the second element is an atm-struct.
;;; - A parsed-bond-starlet is a list, the first element being
;;;   an integer which denotes the number of the center-atom from
;;;   which bond(s) emanate in a star-like fashion, and the
;;;   second element is a list of integers which denote the other ends
;;;   of the bond(s). All these numbers refer to the corresponding
;;;   parsed-atm-vector.

(setf *print-pretty* t)

(defun get-parsed-atm-number (parsed-atm)
  (elt parsed-atm 0))

(defun get-parsed-atm (parsed-atm)
  (elt parsed-atm 1))

(defun get-center-atm (bond-starlet)
  (elt bond-starlet 0))

(defun get-bond-ends (bond-starlet)
  (elt bond-starlet 1))

(defsetf get-bond-ends set)

;;; In make-moiety-type, name is a string naming the moiety,
;;; raw-format is the output from parsing a molecule-describing file,
;;; coordinate-atm-numbers is a list of three atom numbers from
;;; the (PDB) file, those that define the moiety's
;;; internal coordinate system.
;;; The site-num-cap-nums-list is a list containing lists, each of
;;; which begins with the official number of an extension site,
;;; (from 0 to (- n 1) for a moiety with n sites)
;;; and continues with the (PDB) numbers of all capping atoms for
;;; that site. If a site were to have no capping atoms, it would
;;; still be included as a (site-number) entry.

(defun create-moiety-type (&key name
                                raw-format
                                coordinate-atms
                                (sites '()))
  ;; Error check has to be extended to make sure that atoms-to-drop
  ;;   are not connected to any kind of bonds: no dangling bonds,
  ;;   and no unconnected atoms!

  ;; Real work:
  (let* ((sites (copy-tree sites))
         (parsed-atm-vector (copy-seq (first raw-format)))
         (parsed-bond-starlet-vector (copy-seq (second raw-format)))
         (ordered-parsed-atm-vector
          (order-parsed-atm-vector parsed-atm-vector
                                   coordinate-atms))
         (normordered-parsed-atm-vector
          (normalize-parsed-atm-vector ordered-parsed-atm-vector)))
    ;; Change all numbers to atm-array indices:
    (multiple-value-bind (unordered-bond-starlet-vector
                          sites)
                         (renumber-atms
                          normordered-parsed-atm-vector
                          parsed-bond-starlet-vector
                          sites)
      (make-moiety-type
    ;; :number   **gets assigned at the time of entry into the mlib** 
       :name        name
       :atm-array   (map 'vector #'(lambda (parsed-atm)
                                     (get-parsed-atm parsed-atm))
                         normordered-parsed-atm-vector)
       :bond-array  (order&complete-bond-array
                     unordered-bond-starlet-vector)
       :sites       (make-site-array sites)
       ))))

#|
;; Test with bogus sites, caps:
(setf sample-moiety-type
      (create-moiety-type
       :name            "norbornyl"
       :raw-format      raw-format
       :coordinate-atms '(1 7 4)
       :sites           '((0
                           (5)
                           (16 17))
                          (1
                           (6 7)
                           (12 13 15 18))
                          (2
                           (3)
                           ()))
       ))

;; More abstractly:
(create-moiety-type
 :name            "name-string"
 :pdb-parse       pdb-parsing-result
 :coordinate-atms '(,first ,second ,third)
 :sites           '((,number0
                     (,one-or-more-root-atoms)
                     (,one-or-more-capping-atoms))
                    (,number1
                     (,one-or-more-root-atoms)
                     (,one-or-more-capping-atoms)))
 )
|#

(defun make-site-array (sites)
  (map 'vector #'(lambda (site)
                   (make-site :root-atms   (second site)
                              :cap-atms  (third site)
                              :link-types '()))
       sites))

;;; Variables as in make-moiety-type; returns a vector
;;; of atms with the three coordinate atoms first, and the
;;; hydrogen atoms last.
(defun order-parsed-atm-vector (parsed-atm-vector
                                   coordinate-atm-numbers)
  (let ((coordinate-vector (make-array 0 :adjustable t
                                       :fill-pointer 0))
        (temp-vector (copy-seq parsed-atm-vector)))
    ;; Extract the coordinate-defining parsed-atms.
    (dosequence (n coordinate-atm-numbers)
      (vector-push-extend
       (find-if #'(lambda (parsed-atm)
                    (= n (get-parsed-atm-number parsed-atm)))
                parsed-atm-vector)
       coordinate-vector)
      (setf temp-vector
            (remove-if #'(lambda (parsed-atm)
                           (= n (get-parsed-atm-number parsed-atm)))
                       temp-vector)))
    ;; Put them first, H atms last, in coordinate-vector
    (concatenate 'vector
     coordinate-vector
     (stable-sort temp-vector       ; H at end aids the
           #'(lambda (a b)          ;   H-first clash search
               (declare (ignore b)); to not make compiler unhappy
               (not (atm-type=H-p (get-parsed-atm a))))))))

;;; Make and apply a mapping from pdb to reordered numbers;
;;; return a renumbered bond-starlet vector corresponding to array
;;; values.
(defun renumber-atms (parsed-atm-vector
                      parsed-bond-starlet-vector
                      sites)
  ;; Make the mapping array.
  (let* ((n (largest-parsed-atm-number parsed-atm-vector))
         (mapping-array (make-array (+ 1 n))))
    (dotimes (i (length parsed-atm-vector))
      (setf (aref mapping-array
                  (get-parsed-atm-number (elt parsed-atm-vector i)))
            i))
    (flet ((update-list
            (input-list)
            (maplist
             #'(lambda (remainder)
                       (aref mapping-array (first remainder)))
             input-list)))
      ;; Apply mapping to convert PDB numbers.
      (values
       (map 'vector #'(lambda (bond-starlet)
                        (list
                         (car (update-list
                               (list
                                (get-center-atm
                                 bond-starlet))))
                         (update-list (get-bond-ends
                                       bond-starlet))))
            parsed-bond-starlet-vector)
       (maplist
        #'(lambda (remainder)
            (list (first (first remainder))
                  (update-list (second (first remainder)))
                  (update-list (third (first remainder)))))
        sites)))))

;;; Auxilliary to renumber-atms.
(defun largest-parsed-atm-number (parsed-atm-vector)
  (let ((biggest 0))
    (dolist (parsed-atm (coerce parsed-atm-vector 'list))
      (let ((number (get-parsed-atm-number parsed-atm)))
        (when (> number biggest)
          (setf biggest number))))
    biggest))

(defun order&complete-bond-array (unordered-bond-starlet-vector)
  "unordered-bond-starlet-vector
Makes the bond-array that goes into the moiety-type. The bond-starlets
are ordered so that their center-atms correspond to the position of
atms in the atm-array. Missing bonds are reconstructed (certain
PDB-generators have the idiosyncrasy (or idiocy ?) to leave away
all bonds originating from hydrogens)."
  (let ((bond-array (make-array 1 :initial-element '()
                                  :adjustable t)))
    (flet ((insert-with-expand
            (index end-atm)
            (if (< (length bond-array) (+ 1 index))
              (setf bond-array
                    (adjust-array bond-array (+ 1 index)
                                  :initial-element '())))
            (let ((end-atm-list  (elt bond-array index)))
              (if (not (member end-atm end-atm-list))
                (progn (push end-atm end-atm-list)
                       (setf (elt bond-array index)
                             (sort end-atm-list #'<)))))))
      (do ((position 0 (incf position)))
          ((>= position (length unordered-bond-starlet-vector))
            bond-array)
        (let* ((bond-starlet
                (elt unordered-bond-starlet-vector position))
               (ba-index (get-center-atm bond-starlet)))
          (dolist (end-atm (get-bond-ends bond-starlet))
            (insert-with-expand ba-index end-atm)
            (insert-with-expand end-atm ba-index)))))))

