;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    this file is part of CavityStuffer - a molecular polymer design tool
;;;                         -------------
;;; copyright (c) 1993 by Institute for Molecular Manufacturing, Palo Alto, CA
;;; copyright (c) 1994 and beyond: M. Krummenacker (KR) kr@n-a-n-o.com
;;;
;;; -This program is free software; you can redistribute it and/or modify
;;;   it under the terms of the GNU General Public License as published by
;;;   the Free Software Foundation.
;;; -This program is distributed in the hope that it will be useful, but WITHOUT
;;;   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;;;   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU GPL for more details.
;;; -You should have received a copy of the GNU General Public License along with
;;;   this program (the file "COPYING-2.0"); if not, write to the Free Software
;;;   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package "CAVSTUF")

(setf *print-pretty* t)

(defconstant Whitespace-filter '(#\Space #\Tab #\Newline)
  "for string-trim functions")

#|kr96-12-31: moved to utils.lisp
(defun dump-hash-table (hash-table)
  (let ((dump-list (make-array 0 :adjustable t
                                 :element-type 'cons
                                 :fill-pointer 0)))
       (maphash #'(lambda (key val)
                    (vector-push-extend (list key val)
                                        dump-list))
                hash-table)
       (coerce dump-list 'list)))
|#

(defun reconstruct-hash-table (dump-list &key (test #'eql))
  (let ((hash-table (make-hash-table :test test))) ; was #'equal
    (dosequence (item dump-list hash-table)
                (setf (gethash (first item) hash-table)
                      (second item)))))

(defun save-object-in-file (object file-name)
  "object
   file-name ;string
Saves the object in a file of the specified name"
  (with-open-file (out-filestream 
                     file-name
                     :direction :output
                     :if-does-not-exist :create
                     :if-exists :supersede)
    (pprint object out-filestream)
    (pprint (concatenate 'string
                         "the object has been saved in the file: "
                         (coerce '(#\") 'string)
                         file-name
                         (coerce '(#\") 'string)))))

(defun read-object-from-file (file-name)
  "file-name ;string
Reads first item in the specified file and returns this object"
  (with-open-file (in-filestream 
                     file-name
                     :direction :input)
    (read in-filestream)))

(defun save-symbol-value-in-file (object
                       &optional (file-name (write-to-string object)))
  "object ;having a name that could translate into a filename
   &optional file-name ;string to override default filename
Saves the symbol-value of a quoted object in a file of the same name,
unless &optional file-name is specified"
    (save-object-in-file (symbol-value object) file-name))

(defun read-symbol-value-from-file (object
                       &optional (file-name (write-to-string object)))
  "object ;having a name that corresponds to a filename
   &optional file-name ;string to override default filename
Reads first item in a file and produces an object of the same name,
unless &optional file-name is specified"
    (setf (symbol-value object) (read-object-from-file file-name))
    (values (symbol-value object) object))

#|
(save-symbol-value-in-file 'pdb-parse)
(read-symbol-value-from-file 'pdb-parse)
(save-symbol-value-in-file 'pdb-parse "test")
(read-symbol-value-from-file 'pdb-parse "test")

(save-object-in-file '((12 34) (23 45)) "test")
(read-symbol-value-from-file 'pdb-parse "test")
(read-symbol-value-from-file 'pdb-parse)
|#


(defvar Active-mlib nil
  "Denotes the currently active moiety-library")

(defun get-current-mlib ()
  "no parameters
Obtains the currently active moiety-library"
  (values Active-mlib
          (symbol-value Active-mlib)))

(defun init-library-numbers ()
  (setf (get (get-current-mlib) 'library-numbers)
        (make-array 0 :adjustable t
                      :initial-element nil)))

(defun vector-insert-extend (new-element
                                vector
                                &optional (position (length vector)))
  "new-element
   vector              ; has to be :adjustable, but no :fillpointer!!
   &optional position  ; default if not supplied is the length of the vector
Takes the new-element and inserts it into the vector at the specified
position (counting from zero, as usual). Adjusts the length of the vector
beforehand if the position lies outside the previous length."
  (if (>= position (length vector))
    (setf vector (adjust-array vector
                               (+ 1 position)
                               :initial-element nil)))
  (setf (elt vector position) new-element))

(defun reconstruct-adjustable-vector (recalcitrant-vector)
  (let ((reasonably-flexible-vector (make-array 0
                                                :adjustable t
                                                :initial-element nil)))
    (dosequence (element recalcitrant-vector reasonably-flexible-vector)
                (vector-insert-extend element reasonably-flexible-vector))))

(defun next-free-library-number ()
  (let ((hole (position-if #'(lambda (entry)
                               (eql entry nil))
                           (get (get-current-mlib) 'library-numbers))))
    (if (eql hole nil)
      (length (get (get-current-mlib) 'library-numbers))
      hole)))

#|;;;kr98-04-19: not quite sure when this here was introduced,
;;; and which version is more correct. in any case, it seems to
;;; work with the more cumbersome method below
(defun insert-into-library-numbers (item position)
  (vector-insert-extend item
                        (get (get-current-mlib) 'library-numbers)
                        position))
|#

(defun insert-into-library-numbers (item position)
  (if (>= position (length (get (get-current-mlib) 'library-numbers)))
    (setf (get (get-current-mlib) 'library-numbers)
          (adjust-array (get (get-current-mlib) 'library-numbers)
                        (+ 1 position)
                        :initial-element nil)))
  (setf (elt (get (get-current-mlib) 'library-numbers) position)
        (copy-seq item)))

(defun moiety-type-library-number-entry (moiety-type)
  "moiety-type
Not only assigns a unique number to the moiety-type to be entered into
the mlib, but also burns in that number into the moiety-type itself"
  (let ((position (next-free-library-number)))
    (insert-into-library-numbers (moiety-type-name moiety-type) position)
    (setf (moiety-type-number moiety-type) position)))

(defun link-type-library-number-entry (link-type)
  "link-type
Not only assigns a unique number to the link-type to be entered into
the mlib, but also burns in that number into the link-type itself"
  (let ((position (next-free-library-number)))
    (insert-into-library-numbers (link-type-name link-type) position)
    (setf (link-type-number link-type) position)))

(defun delete-library-number (position)
  (insert-into-library-numbers nil position))

(defun construct-lib-file-name (name extension)
  "name         ;string
   extension    ;string (without the dot)
Constructs the full pathname of a library file as a string"
  (namestring 
   (make-pathname :directory Lib-directory
                  :name
                  (concatenate 'string name "." extension)
                  )))

(defun set-current-mlib-to (library-object)
  "library-object ;having a name to which a corresponding file exists
Activates a moiety-library upon which all functions operate,
which involves reading from the file"
  (setf (symbol-value library-object) nil)
  (setf Active-mlib library-object)
  (let* ((name-with-ext (construct-lib-file-name
                         (write-to-string library-object)
                         "mlib"))
         (file-contents (read-object-from-file name-with-ext)))
    (setf (get (get-current-mlib) 'library-numbers)
          (reconstruct-adjustable-vector (car file-contents)))
    (setf (symbol-value (get-current-mlib))
          (reconstruct-hash-table (cdr file-contents)))
    (values (get-current-mlib)
            (symbol-value (get-current-mlib)))))

(defun save-mlib ()
  "no parameters
Saves the contents of the currently active library in
the corresponding file"
  (let ((name-with-ext (construct-lib-file-name
                                    (write-to-string (get-current-mlib))
                                    "mlib")))
    (save-object-in-file (cons
                          (get (get-current-mlib) 'library-numbers)
                          (dump-hash-table
                           (symbol-value (get-current-mlib))))
                         name-with-ext)
    (values (get-current-mlib)
            (symbol-value (get-current-mlib)))))
  
(defun init-mlib (library-object &key (test #'eql))
  "library-object  ;having a name that could translate into a filename
Sets up a new moiety-library including the corresponding file on disk"
  (setf (symbol-value library-object) (make-hash-table :test test)) ; was #'equal
  (setf Active-mlib library-object)
  (init-library-numbers)
  (values))

(defun add-to-mlib (moiety-type)
  "moiety-type
Adds the moiety-type to the moiety-library"
  (let* ((moiety-name (moiety-type-name moiety-type))
         (previous? (multiple-value-bind
                      (value entry-found?)
                      (retrieve-from-mlib moiety-name)
                      (declare (ignore value))
                      entry-found?)))
    (if previous?
      (progn (delete-library-number (moiety-type-number
                                     (retrieve-from-mlib moiety-name)))
             (pprint (concatenate 'string
                                  "the entry: "
                                  (coerce '(#\") 'string)
                                  moiety-name
                                  (coerce '(#\") 'string)
                                  " was overwritten!"))))
    (moiety-type-library-number-entry moiety-type)
    (setf (gethash
           ;;kr96-07-06: make symbols instead of strings.
           (intern (string-upcase moiety-name))
           (symbol-value (get-current-mlib)))
          moiety-type)))

(defun add-link-type-to-mlib (link-type)
  "link-type
Adds the link-type to the moiety-library"
  (let* ((link-name (link-type-name link-type))
         (previous? (multiple-value-bind
                      (value entry-found?)
                      (retrieve-from-mlib link-name)
                      (declare (ignore value))
                      entry-found?)))
    (if previous?
      (progn (delete-library-number (link-type-number
                                     (retrieve-from-mlib link-name)))
             (pprint (concatenate 'string
                                  "the entry: "
                                  (coerce '(#\") 'string)
                                  link-name
                                  (coerce '(#\") 'string)
                                  " was overwritten!"))))
    (link-type-library-number-entry link-type)
    (setf (gethash
           ;;kr96-07-06: make symbols instead of strings.
           (intern (string-upcase link-name))
           (symbol-value (get-current-mlib)))
          link-type)
    (enter-link-type-into-parent-site link-type)))

;;; this is probably not used much ...
;;;
(defun delete-from-mlib (moiety-type)
  "moiety-type
Removes the moiety-type from the moiety-library"
  (delete-library-number (moiety-type-number moiety-type))
  (alarm-link-types-if-parent-deleted moiety-type)
  (remhash
   ;;kr96-07-06: make symbols instead of strings.
   (intern (string-upcase (moiety-type-name moiety-type)))
   (symbol-value (get-current-mlib)))
  (save-mlib)) ; ?? why save here ? could be dangerous, without any backups.

(defun delete-link-type-from-mlib (link-type)
  "moiety-type
Removes the moiety-type from the moiety-library"
  (delete-library-number (link-type-number link-type))
  (delete-link-type-in-parent-site link-type)
  (remhash
   ;;kr96-07-06: make symbols instead of strings.
   (intern (string-upcase (link-type-name link-type)))
   (symbol-value (get-current-mlib)))
  (save-mlib))

(defun retrieve-from-mlib (moiety-type-name)
  "moiety-type-name  ;the name string
Hands back the moiety-type from the moiety-library"
  (gethash
   ;;kr96-07-06: make symbols instead of strings.
   (intern (string-upcase moiety-type-name))
   (symbol-value (get-current-mlib))))


#|
(init-mlib 'experimental)
(get-current-mlib)
(set-current-mlib-to 'experimental)
(add-to-mlib sample-moiety-type)
(delete-from-mlib sample-moiety-type)
(retrieve-from-mlib "tetrahedrane")
(retrieve-from-mlib "norbornoxy")
(save-mlib)
(dump-hash-table (symbol-value (get-current-mlib)))
|#
