;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    this file is part of CavityStuffer - a molecular polymer design tool
;;;                         -------------
;;; copyright (c) 1993 by Institute for Molecular Manufacturing, Palo Alto, CA
;;; copyright (c) 1994 and beyond: M. Krummenacker (KR) kr@n-a-n-o.com
;;;
;;; -This program is free software; you can redistribute it and/or modify
;;;   it under the terms of the GNU General Public License as published by
;;;   the Free Software Foundation.
;;; -This program is distributed in the hope that it will be useful, but WITHOUT
;;;   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;;;   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU GPL for more details.
;;; -You should have received a copy of the GNU General Public License along with
;;;   this program (the file "COPYING-2.0"); if not, write to the Free Software
;;;   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Historical Note: the first version of this file was made by K. E. Drexler.
;;;                  subsequent modifications by KR.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package "CAVSTUF")

;;; Requires Matrix functions

;;; Take a set of three points and return a matrix
;;; that maps points into a coordinate system in which
;;; the first point in the given set is the origin, the 
;;; vector from the first to the second point defines the
;;; x axis, and the vector from the first to the third
;;; point defines the xy plane.
(defun normalizing-matrix (point1 point2 point3)
  (let* ((result-matrix (new-matrix))
         (vector1 (vv- point2 point1))
         (vector2 (vv- point3 point1))
         (unit-x (unit vector1))
         (unit-y (orthogonal-unit vector1 vector2))
         (unit-z (cross unit-x unit-y)))
    (dotimes (j 3)
      (setf (aref result-matrix j 3)
            0.0))
    (dotimes (i 3)
      (setf (aref result-matrix 0 i)
            (aref unit-x i)))
    (dotimes (i 3)
      (setf (aref result-matrix 1 i)
            (aref unit-y i)))
    (dotimes (i 3)
      (setf (aref result-matrix 2 i)
            (aref unit-z i)))
    (let ((point-image (mv* result-matrix point1)))
      (dotimes (j 3)
        (setf (aref result-matrix j 3)
              (- (aref point-image j))))
      result-matrix)))
#|
(normalizing-matrix #(0 0 0) #(1 0 0) #(0 1 0))
(normalizing-matrix #(1 1 1) #(2 1 1) #(1 2 1))
(normalizing-matrix #(0 1 0) #(2 0 0) #(-6 -2 0))
(mv* (normalizing-matrix #(1 3 1) #(4 1 1) #(1 2 1))
     #(1 3 1)
     )
|#
(setf *print-pretty* t)
;;; Take a vector of vector representing
;;; points in space, and return a vector of parsed-atms
;;; in a coordinate system of the sort defined above,
;;; taking the first three points as the origin, etc.
(defun normalize-parsed-atm-vector (ordered-parsed-atm-vector)
  (let ((normalizer (normalizing-matrix
                     (atm-xyz (get-parsed-atm
                               (elt ordered-parsed-atm-vector 0)))
                     (atm-xyz (get-parsed-atm
                               (elt ordered-parsed-atm-vector 1)))
                     (atm-xyz (get-parsed-atm
                               (elt ordered-parsed-atm-vector 2)))
                                        )))
    (map 'vector #'(lambda (parsed-atm)
                     (let ((norm-atm (copy-atm (get-parsed-atm parsed-atm))))
                       (setf (atm-xyz norm-atm)
                             (mv* normalizer
                                  (atm-xyz norm-atm)))
                       (list (get-parsed-atm-number parsed-atm) norm-atm)))
         ordered-parsed-atm-vector)))

;;; Take a list of coordinate vectors representing
;;; points in space, and return a list of vectors
;;; in a coordinate system of the sort defined above,
;;; taking the first three points as the origin, etc.
(defun normalize-points (list-of-points)
  (let ((normalizer (normalizing-matrix (first list-of-points)
                                        (second list-of-points)
                                        (third list-of-points)
                                        )))
    (mapcar #'(lambda (point) (mv* normalizer point))
            list-of-points)))
#|
(let* ((a (list #(1 2 3) #(3 7 5) #(2 3 6) ; mess of points
                #(4 7 6) #(1 7 5) #(7 1 4)))
       (b (normalize-points a)))
  (print b)
  (list            ; check for rigid body translation
   (- (v-distance (elt a 2) (elt a 5))
      (v-distance (elt b 2) (elt b 5)))
   (- (v-distance (elt a 3) (elt a 4))
      (v-distance (elt b 3) (elt b 4)))
   (- (v-distance (elt a 5) (elt a 4))
      (v-distance (elt b 5) (elt b 4)))
   (- (v-distance (elt a 0) (elt a 5))
      (v-distance (elt b 0) (elt b 5)))))
|#

;;; Take two lists of three or more points and return a matrix
;;; that maps points defined relative to a given coordinate
;;; system into points specified relative to the coordinate
;;; system of the second of the two lists, but described
;;; in the coordinate system of the first of the two lists
;;; In application: atom positions defined in child-moiety
;;; coordinates will, when multiplied by the given matrix,
;;; be described in the coordinates of the parent-moiety.
(defun mapping-matrix (point-list1 point-list2)
  (let* ((normalizer1 (normalizing-matrix (first point-list1)
                                          (second point-list1)
                                          (third point-list1)))
         (point21 (mv* normalizer1 (first point-list2)))
         (point22 (mv* normalizer1 (second point-list2)))
         (point23 (mv* normalizer1 (third point-list2)))
         (normalizer2 (normalizing-matrix point21 point22 point23))
         (result-matrix (new-matrix)))
    (dotimes (i 3)
      (setf (aref result-matrix i 0)
            (aref normalizer2   0 i)))
    (dotimes (i 3)
      (setf (aref result-matrix i 1)
            (aref normalizer2   1 i)))
    (dotimes (i 3)
      (setf (aref result-matrix i 2)
            (aref normalizer2   2 i)))
    (dotimes (j 3)
      (setf (aref result-matrix j 3)
            (aref point21 j)))
    result-matrix))
#|

;; Same points three ways, showing different mapping pathways:

(let* ((a (list #(0 0 0) #(3 7 5) #(2 3 6) #(2 1 0)))     ; mess of points
       (b (list #(1 2 3) #(1 7 5) #(7 1 4) #(1 2 0)))     ; mess of points
       (c (list #(5 3 2) #(7 2 3) #(1 3 2) #(6 3 2)))     ; mess of points
       (d (normalize-points (append a b c)))        ; all in a coordinates
       (e (normalize-points b))                ; b points in b coordinates
       (f (normalize-points c))                ; c points in c coordinates
       (g (mapping-matrix b c))                ; mapping of c coord. into b
       (h (mapping-matrix a b))                ; mapping of b coord. into a
       (i (mm* h g))                           ; mapping of c coord. into a
       (j (mapcar #'(lambda (point)            ; c points in b coordinates
                       (mv* g point))
                   f)))
  (print (subseq d 8))                ; c points in a coordinates directly
  (print
   (mapcar #'(lambda (point)
               (mv* h point))
           j))                ; c points in a coordinates via two matrices
  (mapcar #'(lambda (point)
              (mv* i point))
          f))                ; c points in a coordinates via single matrix

|#
