;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;    this file is part of CavityStuffer - a molecular polymer design tool
;;;                         -------------
;;; copyright (c) 1993 by Institute for Molecular Manufacturing, Palo Alto, CA
;;; copyright (c) 1994 and beyond: M. Krummenacker (KR) kr@n-a-n-o.com
;;;
;;; -This program is free software; you can redistribute it and/or modify
;;;   it under the terms of the GNU General Public License as published by
;;;   the Free Software Foundation.
;;; -This program is distributed in the hope that it will be useful, but WITHOUT
;;;   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;;;   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU GPL for more details.
;;; -You should have received a copy of the GNU General Public License along with
;;;   this program (the file "COPYING-2.0"); if not, write to the Free Software
;;;   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package "CAVSTUF")


(defun dump-tree (tree &optional (including-frame? nil))
  "tree                        ;usually would point to the frame-moiety
   &optional including-frame?  ;defaults to leaving frame away
Generates the raw-format description of the supplied tree, consisting of
a vector of parsed-atms and a vector of parsed-bond-starlets 
which correspond in their sequence
to the atm-array, the numbering of the parsed-atms starting at offset"
  (let ((parsed-atm-vector  (make-array 0 :adjustable t
                                        :fill-pointer 0))
        (parsed-bond-starlet-vector  (make-array 0 :adjustable t
                                                   :fill-pointer 0))
        (atm-offset 0))
    (labels ((append-to-parsed-atm-vector (atm-array
                                               offset
                                               &optional (cap-atms '()))
  "atm-array
   offset
   &optional cap-atms  ; list of indices of cap-atms in 'used sites
Generate a vector of parsed-atms, the sequence of which corresponds
to the atm-array, the numbering of the parsed-atms starting at offset.
Suppress the output of cap-atms.
Returns the offset which has been incremented by the number of added atms."
  (let ((this-many (length atm-array)))
    (dotimes (index this-many)
      (unless (member index cap-atms)
        (vector-push-extend (list (+ index offset) (elt atm-array index))
                            parsed-atm-vector)))
    (values (+ this-many offset))))
             (append-to-parsed-bond-starlet-vector (bond-array
                                                          offset
                                                          &optional
                                                          (cap-atms '())
                                                          (root-atms '()))
  "bond-array
   offset
   &optional cap-atms  ; list of indices of cap-atms in 'used sites
   &optional root-atms ; list of indices of root-atms in 'used sites
Generate a vector of parsed-bond-starlets which correspond in their
sequence to the bond-array, the numbering of the center-atms starting at offset
and the end-atms corrected with offset.
Suppress the output of bonds involving cap-atms or root-atms as a center-atm."
  (dotimes (index (length bond-array) parsed-bond-starlet-vector)
    (unless (or (member index cap-atms)             ;;suppress cap-atms
                (member index root-atms))           ;;suppress root-atms
      (let ((end-atms (elt bond-array index)))
        (vector-push-extend (list (+ index offset)
                                  (map 'list
                                       #'(lambda (end-atm)
                                           (+ end-atm offset))
                                       end-atms))
                            parsed-bond-starlet-vector)))))
             (append-dumped-node (tree atm-offset cap-atms root-atms)
  "tree
   atm-offset
   cap-atms
   root-atms
Appends the atms and bonds of a moiety-'node' to the respective vectors.
Leaves out the capping atms (with the respective bonds) of sites
which are 'used. Returns the updated value of atm-offset."
  (append-to-parsed-bond-starlet-vector
   (moiety-type-bond-array (moiety-moiety-type tree))
   atm-offset
   cap-atms
   root-atms)
  (values (append-to-parsed-atm-vector
           (moiety-atm-array tree)
           atm-offset
           cap-atms)))
             (append-dumped-linking-bonds (subtree
                                               cap-atms
                                               parent-atm-offset
                                               link-type)
  "subtree              ; points to the child-moiety
   cap-atms             ; child's cap-atms
   parent-atm-offset    ; the atm-offset when the parent was processed
   link-type            ; link-type by which subtree is hooked up to parent
Appends the bonds between two moiety-'nodes' to the parsed-bond-vector.
This routine has a retrospective behaviour; it links the current subtree-'node'
to its parent-'node'. These bonds constitute the actual linkages of
the (independent) moieties in the stuffing-tree."
  (let* ((parent-bonds (link-type-parent-bonds link-type))
         (child-bonds (link-type-child-bonds link-type))
         (parent-type (retrieve-from-mlib (link-type-parent link-type)))
         (parent-bond-array (moiety-type-bond-array parent-type))
         (child-bond-array (moiety-type-bond-array
                             (moiety-moiety-type subtree)))
         (parent-cap-atms (site-cap-atms (elt (moiety-type-sites parent-type)
                                              (link-type-parent-site
                                               link-type)))))
    (dosequence (bond-starlet parent-bonds)
                (let ((center-atm (get-center-atm bond-starlet)))
                  (vector-push-extend
                   (list (+ center-atm parent-atm-offset)
                         (concatenate
                          'list
                          ;; the parents root-atms but without the cap-atms
                          (map 'list #'(lambda (end-atm)
                                         (+ end-atm parent-atm-offset))
                               (remove-if #'(lambda (end-atm)
                                              (member end-atm
                                                      parent-cap-atms))
                                          (elt parent-bond-array
                                               center-atm)))
                          ;; instead, add the root-atms of the child
                          (map 'list #'(lambda (end-atm)
                                         (+ end-atm atm-offset))
                               (get-bond-ends bond-starlet))))
                   parsed-bond-starlet-vector)))
    (dosequence (bond-starlet child-bonds)
                (let ((center-atm (get-center-atm bond-starlet)))
                  (vector-push-extend
                   (list (+ center-atm atm-offset)
                         (concatenate
                          'list
                          ;; the childs root-atms but without the cap-atms
                          (map 'list #'(lambda (end-atm)
                                         (+ end-atm atm-offset))
                               (remove-if #'(lambda (end-atm)
                                              (member end-atm
                                                      cap-atms))
                                          (elt child-bond-array
                                               center-atm)))
                          ;; instead, add the root-atms of the parent
                          (map 'list #'(lambda (end-atm)
                                         (+ end-atm parent-atm-offset))
                               (get-bond-ends bond-starlet))))
                   parsed-bond-starlet-vector)))))
             (get-cap&root-atms (moiety)
  "moiety
"
  (let ((cap-atms '())
        (root-atms '())
        (sites (moiety-type-sites
                (moiety-moiety-type moiety))))
    (dotimes (site-index (length sites) (values cap-atms root-atms))
      (when (eql (elt (moiety-site-status moiety) site-index)
                 'used)
        (setf cap-atms (append cap-atms (site-cap-atms
                                         (elt sites site-index))))
        (setf root-atms (append root-atms (site-root-atms
                                         (elt sites site-index))))))))
             (dump-the-subtrees (parent-tree parent-atm-offset stem-site)
  "parent-tree
   parent-atm-offset    ; needed to get the links btwn. moieties right
   stem-site            ; site-index by which the moiety pointed to by
                        ;   parent-tree is anchored higher up
Appends the atms and bonds of the moiety-'nodes' of the subtrees to the
respective vectors.
This routine will result in strict depth first traversion of the tree."
  (let ((site-status (moiety-site-status parent-tree)))
    (dotimes (site-number (length site-status) (values))
      ;; if no subtree exists, skip on to the next possibility
      (when (eql (elt site-status site-number) 'used)
          ;; don't follow up on the stem-site: that's where you came from!
          (unless (equal site-number stem-site)
        (let* ((subtree (elt (moiety-children parent-tree) site-number))
               (link-type (moiety-link-type subtree))
               (new-stem-site (link-type-child-site link-type))
               (previous-atm-offset atm-offset)
               ;;96-6-29: needs to initialized with nil, because of clisp bug.
               (new-atm-offset nil))
          (multiple-value-bind (cap-atms root-atms)
                               (get-cap&root-atms subtree)
            (setf new-atm-offset
                  (append-dumped-node subtree atm-offset cap-atms root-atms))
            (append-dumped-linking-bonds subtree
                                         cap-atms
                                         parent-atm-offset
                                         link-type))
          (setf atm-offset new-atm-offset)
          (dump-the-subtrees subtree
                             previous-atm-offset
                             new-stem-site))))))))

      ;; actual main body of dump-tree:
      ;; dump the atms&bonds of the frame if necessary
      (if including-frame?
        (let ((initial-atm-offset atm-offset))
          (multiple-value-bind (cap-atms root-atms)
                               (get-cap&root-atms tree)
            (setf atm-offset (append-dumped-node tree atm-offset
                                                 cap-atms root-atms)))
          ;;Probably BUG: nil should be replaced by actual stem-site !!!!!!!!!
          (dump-the-subtrees tree initial-atm-offset nil))
        (dump-the-subtrees tree atm-offset nil))

      (generate-raw-format parsed-atm-vector parsed-bond-starlet-vector))))

